import json

import requests
from blessings import Terminal
from pycapo import CapoConfig


class FaultReporter:
    """
    Interface for reporting faults
    """

    def ok(self, message):
        """
        Call to signal that your test passed
        :param message: a message about your test
        """
        pass

    def warn(self, message):
        """
        Call to signal your test produced a warning
        :param message: the message to display about your test
        """
        pass

    def error(self, message):
        """
        Call to signal that your test produced an error
        :param message: the message to display about your test
        """
        pass


class ConsoleReporter(FaultReporter):
    """
    Fault reporter that displays messages on the command line
    """
    def __init__(self):
        self.t = Terminal()

    def ok(self, message):
        r""" Print an '[OK]' message to the terminal.

        :param message: The message to print.
        :return: None
        """
        print('{} {}'.format(self.t.green('[OK]    '), message))

    def warn(self, message):
        """ Print a '[WARN]' message to the terminal.

        :param message: The message to print.
        :return: None
        """
        print('{} {}'.format(self.t.yellow('[WARN]  '), message))

    def error(self, message):
        r""" Print an '[ERROR]' message to the terminal.

        :param message: The message to print.
        :return: None
        """
        print('{} {}'.format(self.t.red('[ERROR] '), message))


class SlackReporter(FaultReporter):
    """
    Fault reporter to send the report back to Slack
    """

    def __init__(self):
        self.results = []
        self.errors = 0
        self.successes = 0
        self.warnings = 0

    def add_message(self, message, color):
        """
        Add a note to the list of notes we're sending to Slack.
        :param message: the message
        :param color:  what to color this message
        """
        self.results.append({'fallback': message, 'text': message, 'color': color})

    def ok(self, message):
        self.successes += 1
        self.add_message(message, "#36a64f")

    def warn(self, message):
        self.warnings += 1
        self.add_message(message, "#f28500")

    def error(self, message):
        self.errors += 1
        self.add_message(message, "#D00")

    def make_summary(self):
        """
        Produces a textual summary of the messages we're tracking for the
        first line of the report
        :return: this summary
        """
        if self.errors == 0 and self.warnings == 0:
            return 'All archive systems are fault-free right now (or someone needs to write a new test)'
        elif self.errors == 0 and self.warnings > 0:
            return 'A few archive systems are generating warnings, but nothing is known to be broken'
        elif self.successes == 0:
            return 'Things are *very* on fire right now!'
        else:
            return 'Several archive systems need attention now!'

    def publish(self):
        """
        Publish the message to Slack using the events destination
        """
        config = CapoConfig().settings('archive.events.slack')
        requests.post(config.webhook, json.dumps({'text': self.make_summary(),
                                                  'attachments': self.results}))


class BulkReporter(FaultReporter):
    def __init__(self):
        self.aggregators = []

    def add_aggregator(self, agg):
        self.aggregators.append(agg)

    def ok(self, message):
        for a in self.aggregators:
            a.ok(message)

    def warn(self, message):
        for a in self.aggregators:
            a.warn(message)

    def error(self, message):
        for a in self.aggregators:
            a.error(message)
